"""
Hata is an async Discord API wrapper written in Python named after Hata no Kokoro.

Why hata
--------

- Multiple simultaneous clients

    Hata can run multiple clients from the same instance without sacrificing performance, all while being easy to code.

- Performant
    
    Fast concurrent code using async/await syntax, cache control, PyPy support and more!

- Newest API features
    
    Whatever Discord decides to release/update/break Hata will support it natively in no time!

- 100% Python

    Completely relies on Python! Easy to read, easy to understand, easy to code.

Usage
-----

The following example answers on `ping` message.

```py
from hata import Client, wait_for_interruption

Nue = Client('TOKEN')

@Nue.events
async def ready(client):
    print(f'{client:f} logged in.')

@Nue.events
async def message_create(client, message):
    if message.author.is_bot:
        return
    
    if message.content == 'ping':
        await client.message_create(message.channel, 'pong')

Nue.start()

wait_for_interruption()
```

An improved example using the `commands` extension to handle common use cases.

```py
from hata import Client, wait_for_interruption

Saki = Client('TOKEN', extensions='commands', prefix='s!')

@Saki.events
async def ready(client):
    print(f'{client:f} logged in.')

@Saki.commands
async def ping(client, message):
    return 'pong'

Saki.start()

wait_for_interruption()
```

Or use slash commands!

```py
from hata import Client, Guild, wait_for_interruption

GUILD = Guild.precreate(guild_id)

Seija = Client('TOKEN', extensions='slash')

@Seija.events
async def ready(client):
    print(f'{client:f} logged in.')

@Seija.interactions(guild=GUILD)
async def ping():
    \"\"\"ping-pong\"\"\"
    return 'pong'

Seija.start()

wait_for_interruption()
```

> Note: You need to restart your client, or the slash command wont show up. If there are more than 50 integrations
> (bots) in a guild, some of the (integrations) bots wont be able to use slash commands. This is currently a Discord
> limitation.

Hata leaves the main thread free, `client.start()` blocks it only till the client logs in (or fails it), although you
can still use the `start_clients()` function, what as it says, starts up all the non-running clients parallelly.

Sometimes leaving the main thread might cause problems when trying to shut down the bot(s). At this case, you might
want to use `wait_for_interruption()`, which disconnects the clients gracefully and closes the event loop on keyboard
interrupt.

We got some tutorials on `github:https://github.com/HuyaneMatsu/hata/tree/master/docs` as well, please check them too!
"""
__version__ = '1.1.129'

from .deprecate_backend import deprecate

from .discord import *
from .ext import *

__all__ = (
    *discord.__all__,
    *ext.__all__,
)

from scarletio import check_satisfaction
check_satisfaction()

deprecate()



__all__ = ('InteractionForm',)

import reprlib

from scarletio import copy_docs

from .components import ComponentBase, ComponentRow, create_component, _debug_component_components, \
    _debug_component_custom_id, _debug_component_title, create_auto_custom_id

class InteractionForm(ComponentBase):
    """
    Form component.
    
    Attributes
    ----------
    components : `None` or `tuple` of ``ComponentBase`` instances
        Stored components.
    custom_id : `None` or `str`
        Custom identifier to match the form data when receiving it's interaction back.
    title : `None` or `str`
        The form's title.
    
    Class Attributes
    ----------------
    type : ``ComponentType`` = `ComponentType.none`
        The component's type.
    """
    __slots__ = ('components', 'custom_id', 'title')
    
    def __new__(cls, title, components, custom_id=None):
        """
        Creates an form interaction instance.
        
        Parameters
        ----------
        title : `None` or `str`
            The form's title.
        components : ``ComponentBase`` instances
            Sub components.
        custom_id : `None` or `str`, Optional
             Custom identifier for the form.
        
        Raises
        ------
        AssertionError
            - If `components` contains a non ``ComponentBase`` instance.
            - If `custom_id` was not given neither as `None` or `str` instance.
            - If `custom_id`'s length is over `100`.
            - If `title`'s length is out of the expected range.
        """
        if __debug__:
            _debug_component_components(components)
            _debug_component_custom_id(custom_id)
            _debug_component_title(title)
        
        # components
        if components:
            components = tuple(
                component if isinstance(component, ComponentRow) else ComponentRow(component)
                for component in components
            )
        else:
            components = None
        
        # custom_id
        if (custom_id is None) and (not custom_id):
            custom_id = create_auto_custom_id()
        
        # title
        # No additional checks
        
        self = object.__new__(cls)
        self.components = components
        self.custom_id = custom_id
        self.title = title
        return self
    
    
    @classmethod
    @copy_docs(ComponentBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        # components
        component_datas = data.get('components', None)
        if (component_datas is None) or (not component_datas):
            components = None
        else:
            components = [create_component(component_data) for component_data in component_datas]
        self.components = components
        
        # custom_id
        self.custom_id = data.get('custom_id', None)
        
        # title
        title = data.get('title', None)
        if (title is not None) and (not title):
            title = None
        self.title = title
        
        return self
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        data = {}
        
        # components
        components = self.components
        if (components is None):
            component_datas = []
        else:
            component_datas = [component.to_data() for component in components]
        data['components'] = component_datas
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            data['custom_id'] = custom_id
        
        
        # title
        title = self.title
        if (title is not None):
            data['title'] = title
        
        return data
    

    @copy_docs(ComponentBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # System fields : custom_id
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id=')
            repr_parts.append(reprlib.repr(custom_id))
        
        # Text fields : label & placeholder
        
        # title
        title = self.title
        if (title is not None):
            repr_parts.append(', title=')
            repr_parts.append(reprlib.repr(title))
        
        # sub-component fields : components
        
        # components
        repr_parts.append(', components=')
        components = self.components
        if (components is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(components)
            
            while True:
                component = components[index]
                index += 1
                
                repr_parts.append(repr(component))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)


    @copy_docs(ComponentBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # components
        components = self.components
        if (components is not None):
            components = tuple(component.copy() for component in self.components)
        
        new.components = components
        
        # custom_id
        new.custom_id = self.custom_id
        
        # title
        new.title = self.title
        
        return new
    
    
    def copy_with(self, **kwargs):
        """
        Copies the component and modifies the created one with the given parameters.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            Keyword parameters referencing attributes.
        
        Other Parameters
        ----------------
        components : `iterable` of ``ComponentBase`` instances, Optional (Keyword only)
            Sub components.
        
        custom_id : `None` or `str`, Optional (Keyword only)
            Custom identifier to detect which button was clicked by the user.
        
        title : `None` or `str`, Optional (Keyword only)
            The form's title.
        
        Returns
        -------
        new : ``InteractionForm``
        """
        # components
        try:
            components = kwargs.pop('components')
        except KeyError:
            components = self.components
        else:
            components = tuple(components)
            
            if __debug__:
                _debug_component_components(components)
            
            if components:
                components = tuple(
                    component if isinstance(component, ComponentRow) else ComponentRow(component)
                    for component in components
                )
            else:
                components = None
        
        # custom_id
        try:
            custom_id = kwargs.pop('custom_id')
        except KeyError:
            custom_id = self.custom_id
        else:
            if __debug__:
                _debug_component_custom_id(custom_id)
            
            if (custom_id is None) or (not custom_id):
                custom_id = create_auto_custom_id()
        
        # title
        try:
            title = kwargs.pop('title')
        except KeyError:
            title = self.title
            
            if __debug__:
                _debug_component_title(title)
        
        
        new = object.__new__(type(self))
        new.components = components
        new.custom_id = custom_id
        new.title = title
        return new
    
    
    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        
        # components
        if self.components != other.components:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # title
        if self.title != other.title:
            return False
        
        return True
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        # type
        hash_value = self.type.value
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= len(components)<<12
            for component in components:
                hash_value ^= hash(component)
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # title
        title = self.title
        if (title is not None):
            hash_value ^= hash(title)
        
        return hash_value
    
    
    @copy_docs(ComponentBase._iter_components)
    def _iter_components(self):
        yield self
        
        components = self.components
        if (components is not None):
            for component in components:
                yield from component._iter_components()
    
    
    @copy_docs(ComponentBase._replace_direct_sub_components)
    def _replace_direct_sub_components(self, relation):
        components = self.components
        if (components is not None):
            self.components = tuple(relation.get(component, component) for component in components)
    
    
    @copy_docs(ComponentBase._iter_direct_sub_components)
    def _iter_direct_sub_components(self):
        components = self.components
        if (components is not None):
            yield from components



__all__ = ()

JSON_KEY_BOT_INFO_ID = 'id'
JSON_KEY_BOT_INFO_NAME = 'username'
JSON_KEY_BOT_INFO_DISCRIMINATOR_STRING = 'discriminator'
JSON_KEY_BOT_INFO_AVATAR_BASE64 = 'avatar'
JSON_KEY_BOT_INFO_BANNER_URL = 'bannerUrl'
JSON_KEY_BOT_INFO_PREFIX = 'prefix'
JSON_KEY_BOT_INFO_SHORT_DESCRIPTION = 'shortdesc'
JSON_KEY_BOT_INFO_LONG_DESCRIPTION = 'longdesc'
JSON_KEY_BOT_INFO_TAG_ARRAY = 'tags'
JSON_KEY_BOT_INFO_WEBSITE_URL = 'website'
JSON_KEY_BOT_INFO_SUPPORT_SERVER_INVITE_URL = 'support'
JSON_KEY_BOT_INFO_GITHUB_URL = 'github'
JSON_KEY_BOT_INFO_OWNER_ID_ARRAY = 'owners'
JSON_KEY_BOT_INFO_FEATURED_GUILD_ID_ARRAY = 'guilds'
JSON_KEY_BOT_INFO_INVITE_URL = 'invite'
JSON_KEY_BOT_INFO_CERTIFIED_AT = 'date'
JSON_KEY_BOT_INFO_IS_CERTIFIED = 'certifiedBot'
JSON_KEY_BOT_INFO_VANITY_URL = 'vanity'
JSON_KEY_BOT_INFO_UPVOTES = 'points'
JSON_KEY_BOT_INFO_UPVOTES_MONTHLY = 'monthlyPoints'
JSON_KEY_BOT_INFO_DONATE_BOT_GUILD_ID = 'donatebotguildid'

JSON_KEY_BOT_STATS_GUILD_COUNT = 'serverCount'
JSON_KEY_BOT_STATS_GUILD_COUNT_PER_SHARD_ARRAY = 'shards'
JSON_KEY_BOT_STATS_SHARD_ID = 'shardId'
JSON_KEY_BOT_STATS_SHARD_COUNT = 'shardCount'

JSON_KEY_POST_BOT_STATS_GUILD_COUNT = 'server_count'
JSON_KEY_POST_BOT_STATS_SHARD_ID = 'shard_id'
JSON_KEY_POST_BOT_STATS_SHARD_COUNT = 'shard_count'

JSON_KEY_USER_INFO_ID = 'id'
JSON_KEY_USER_INFO_NAME = 'username'
JSON_KEY_USER_INFO_DISCRIMINATOR_STRING = 'discriminator'
JSON_KEY_USER_INFO_AVATAR_BASE64 = 'avatar'
JSON_KEY_USER_INFO_BIO = 'bio'
JSON_KEY_USER_INFO_BANNER_URL = 'banner'
JSON_KEY_USER_INFO_CONNECTIONS = 'social'
JSON_KEY_USER_INFO_COLOR = 'color'
JSON_KEY_USER_INFO_IS_SUPPORTER = 'supporter'
JSON_KEY_USER_INFO_IS_CERTIFIED_DEVELOPER = 'certifiedDev'
JSON_KEY_USER_INFO_IS_MODERATOR = 'mod'
JSON_KEY_USER_INFO_IS_WEBSITE_MODERATOR = 'webMod'
JSON_KEY_USER_INFO_IS_ADMIN = 'admin'

JSON_KEY_USER_INFO_CONNECTION_YOUTUBE = 'youtube'
JSON_KEY_USER_INFO_CONNECTION_REDDIT = 'reddit'
JSON_KEY_USER_INFO_CONNECTION_TWITTER = 'twitter'
JSON_KEY_USER_INFO_CONNECTION_INSTAGRAM = 'instagram'
JSON_KEY_USER_INFO_CONNECTION_GITHUB = 'github'

JSON_KEY_WEEKEND_STATUS = 'is_weekend'
JSON_KEY_VOTED = 'voted'
JSON_KEY_EXCEPTION_MESSAGE = 'error'

QUERY_KEY_GET_BOTS_LIMIT = 'limit'
QUERY_KEY_GET_BOTS_OFFSET = 'offset'
QUERY_KEY_GET_BOTS_SORT_BY = 'sort'
QUERY_KEY_GET_BOTS_SEARCH_QUERY = 'search'
QUERY_KEY_GET_BOTS_FIELDS = 'fields'

JSON_KEY_BOTS_QUERY_RESULT_RESULTS = 'results'
JSON_KEY_BOTS_QUERY_RESULT_LIMIT = 'limit'
JSON_KEY_BOTS_QUERY_RESULT_OFFSET = 'offset'
JSON_KEY_BOTS_QUERY_RESULT_COUNT = 'count'
JSON_KEY_BOTS_QUERY_RESULT_TOTAL = 'total'

QUERY_KEY_GET_USER_VOTE_USER_ID = 'userId'

RATE_LIMIT_GLOBAL_SIZE = 100
RATE_LIMIT_GLOBAL_RESET_AFTER = 1.0

RATE_LIMIT_BOTS_SIZE = 60
RATE_LIMIT_BOTS_RESET_AFTER = 60.0

RATE_LIMIT_GLOBAL_DEFAULT_DURATION = 3600.0

JSON_KEY_VOTE_BASE_QUERY = 'query'
JSON_KEY_VOTE_BASE_TYPE = 'type'
JSON_KEY_VOTE_BASE_USER_ID = 'user'

JSON_KEY_BOT_VOTE_BOT_ID = 'bot'
JSON_KEY_BOT_VOTE_IS_WEEKEND = 'isWeekend'

JSON_KEY_GUILD_VOTE_GUILD_ID = 'guild'



__all__ = ()

from scarletio import IgnoreCaseString

AUDIT_LOG_REASON = IgnoreCaseString('X-Audit-Log-Reason')

RATE_LIMIT_REMAINING = IgnoreCaseString('X-RateLimit-Remaining')
RATE_LIMIT_RESET = IgnoreCaseString('X-RateLimit-Reset')
RATE_LIMIT_RESET_AFTER = IgnoreCaseString('X-RateLimit-Reset-After')
RATE_LIMIT_LIMIT = IgnoreCaseString('X-RateLimit-Limit')

# to send
RATE_LIMIT_PRECISION = IgnoreCaseString('X-RateLimit-Precision')
DEBUG_OPTIONS = IgnoreCaseString('X-Debug-Options')



__all__ = ('BotInfo', 'BotStats', 'BotVote', 'BotsQueryResult', 'BriefUserInfo', 'GuildVote', 'UserConnections',
    'UserInfo', 'VoteBase')

from scarletio import copy_docs
from ...discord.utils import timestamp_to_datetime
from ...discord.bases import IconSlot, Slotted
from ...discord.http import urls as module_urls
from ...discord.color import Color

# bot info constants
from .constants import JSON_KEY_BOT_INFO_ID, JSON_KEY_BOT_INFO_NAME, JSON_KEY_BOT_INFO_DISCRIMINATOR_STRING, \
    JSON_KEY_BOT_INFO_AVATAR_BASE64, JSON_KEY_BOT_INFO_BANNER_URL, JSON_KEY_BOT_INFO_PREFIX, \
    JSON_KEY_BOT_INFO_SHORT_DESCRIPTION, JSON_KEY_BOT_INFO_LONG_DESCRIPTION, JSON_KEY_BOT_INFO_TAG_ARRAY, \
    JSON_KEY_BOT_INFO_WEBSITE_URL, JSON_KEY_BOT_INFO_SUPPORT_SERVER_INVITE_URL, JSON_KEY_BOT_INFO_GITHUB_URL, \
    JSON_KEY_BOT_INFO_OWNER_ID_ARRAY, JSON_KEY_BOT_INFO_FEATURED_GUILD_ID_ARRAY, JSON_KEY_BOT_INFO_INVITE_URL, \
    JSON_KEY_BOT_INFO_CERTIFIED_AT, JSON_KEY_BOT_INFO_IS_CERTIFIED, JSON_KEY_BOT_INFO_VANITY_URL, \
    JSON_KEY_BOT_INFO_UPVOTES, JSON_KEY_BOT_INFO_UPVOTES_MONTHLY, JSON_KEY_BOT_INFO_DONATE_BOT_GUILD_ID

# bot stats constants
from .constants import JSON_KEY_BOT_STATS_GUILD_COUNT, JSON_KEY_BOT_STATS_GUILD_COUNT_PER_SHARD_ARRAY, \
    JSON_KEY_BOT_STATS_SHARD_ID, JSON_KEY_BOT_STATS_SHARD_COUNT

# user info constants
from .constants import JSON_KEY_USER_INFO_ID, JSON_KEY_USER_INFO_NAME, JSON_KEY_USER_INFO_DISCRIMINATOR_STRING, \
    JSON_KEY_USER_INFO_AVATAR_BASE64, JSON_KEY_USER_INFO_BIO, JSON_KEY_USER_INFO_BANNER_URL, \
    JSON_KEY_USER_INFO_CONNECTIONS, JSON_KEY_USER_INFO_COLOR, JSON_KEY_USER_INFO_IS_SUPPORTER, \
    JSON_KEY_USER_INFO_IS_CERTIFIED_DEVELOPER, JSON_KEY_USER_INFO_IS_MODERATOR, \
    JSON_KEY_USER_INFO_IS_WEBSITE_MODERATOR, JSON_KEY_USER_INFO_IS_ADMIN

# user connections constants
from .constants import JSON_KEY_USER_INFO_CONNECTION_YOUTUBE, JSON_KEY_USER_INFO_CONNECTION_REDDIT, \
    JSON_KEY_USER_INFO_CONNECTION_TWITTER, JSON_KEY_USER_INFO_CONNECTION_INSTAGRAM, \
    JSON_KEY_USER_INFO_CONNECTION_GITHUB

# bots query result constants
from .constants import JSON_KEY_BOTS_QUERY_RESULT_RESULTS, JSON_KEY_BOTS_QUERY_RESULT_LIMIT, \
    JSON_KEY_BOTS_QUERY_RESULT_OFFSET, JSON_KEY_BOTS_QUERY_RESULT_COUNT, JSON_KEY_BOTS_QUERY_RESULT_TOTAL

# vote base constants
from .constants import JSON_KEY_VOTE_BASE_QUERY, JSON_KEY_VOTE_BASE_TYPE, JSON_KEY_VOTE_BASE_USER_ID

# bot vote constants
from .constants import JSON_KEY_BOT_VOTE_BOT_ID, JSON_KEY_BOT_VOTE_IS_WEEKEND

# guild vote constants
from .constants import JSON_KEY_GUILD_VOTE_GUILD_ID

class BotInfo(metaclass=Slotted):
    """
    Representing information about a bot.
    
    Attributes
    ----------
    avatar_hash : `int`
        The bot's avatar hash.
    avatar_type : ``IconType``
        The bot's avatar's type.
    banner_url : `None` or `str`
        Url for the bot's banner image.
    certified_at : `None` or `datetime`
        When the bot was approved. Set as `None` if was not yet.
    discriminator : `int`
        The bot's discriminator.
    donate_bot_guild_id : `int`
        The guild id for the donate bot setup(?).
    featured_guild_ids : `None` or `tuple` of `int`
        The featured guild's identifiers on the bot's page.
    github_url : `None` or `str`
        Link to the github repo of the bot.
    id : `int`
        The bot's identifier.
    invite_url : `None` or `str`
        Custom bot invite url.
    long_description : `str`
        The long description of the bot.
    name : `str`
        The name of the bot.
    owner_id : `int
        The bot's main owner's identifier.
    owner_ids : `tuple` of `int`
        The bot's owners' identifiers.
    prefix : `str`
        Prefix of the bot.
    short_description : `str`
        The short description of the bot.
    support_server_invite_url : `None` or `str`
        Url to the bot's support server.
    tags : `None` or `tuple` of `str`
        The tags of the bot.
    upvotes : `int`
        The amount of upvotes the bot has.
    upvotes_monthly : `int`
        The amount of upvotes the bot has this month.
    vanity_url : `None` or `str`
        Vanity url of the bot.
    website_url : `None` or `str`
        The website url of the bot.
    """
    __slots__ = ('banner_url', 'certified_at', 'discriminator', 'donate_bot_guild_id', 'featured_guild_ids',
        'github_url', 'id', 'invite_url', 'long_description', 'name', 'owner_id', 'owner_ids', 'prefix',
        'short_description', 'support_server_invite_url', 'tags', 'upvotes', 'upvotes_monthly', 'vanity_url',
        'website_url')
    
    avatar = IconSlot(
        'avatar',
        JSON_KEY_BOT_INFO_AVATAR_BASE64,
        module_urls.user_avatar_url,
        module_urls.user_avatar_url_as,
    )
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new bot info instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Deserialized bot info data.
        
        Returns
        -------
        self : ``BotInfo``
        """
        self = object.__new__(cls)
        
        # avatar_hash & avatar_type
        self._set_avatar(data)
        
        # banner_url
        self.banner_url = data.get(JSON_KEY_BOT_INFO_BANNER_URL, None)
        
        # certified_at
        if data.get(JSON_KEY_BOT_INFO_IS_CERTIFIED, False):
            certified_at = timestamp_to_datetime(data[JSON_KEY_BOT_INFO_CERTIFIED_AT])
        else:
            certified_at = None
        self.certified_at = certified_at
        
        # discriminator
        self.discriminator = int(data[JSON_KEY_BOT_INFO_DISCRIMINATOR_STRING])
        
        # donate_bot_guild_id
        donate_bot_guild_id = data[JSON_KEY_BOT_INFO_DONATE_BOT_GUILD_ID]
        if donate_bot_guild_id:
            donate_bot_guild_id = int(donate_bot_guild_id)
        else:
            donate_bot_guild_id = 0
        self.donate_bot_guild_id = donate_bot_guild_id
        
        # featured_guild_ids
        featured_guild_ids = data.get(JSON_KEY_BOT_INFO_FEATURED_GUILD_ID_ARRAY, None)
        if (featured_guild_ids is None) or (not featured_guild_ids):
            featured_guild_ids = None
        else:
            featured_guild_ids = tuple(sorted(int(guild_id) for guild_id in featured_guild_ids))
        self.featured_guild_ids = featured_guild_ids
        
        # github_url
        self.github_url = data.get(JSON_KEY_BOT_INFO_GITHUB_URL, None)
        
        # id
        self.id = int(data[JSON_KEY_BOT_INFO_ID])
        
        # invite_url
        self.invite_url = data.get(JSON_KEY_BOT_INFO_INVITE_URL, None)
        
        # long_description
        self.long_description = data[JSON_KEY_BOT_INFO_LONG_DESCRIPTION]
        
        # name
        self.name = data[JSON_KEY_BOT_INFO_NAME]
        
        # owner_id & owner_ids
        owner_ids = data[JSON_KEY_BOT_INFO_OWNER_ID_ARRAY]
        self.owner_id = int(owner_ids[0])
        self.owner_ids = tuple(sorted(int(owner_id) for owner_id in owner_ids))
        
        # prefix
        self.prefix = data[JSON_KEY_BOT_INFO_PREFIX]
        
        # short_description
        self.short_description = data[JSON_KEY_BOT_INFO_SHORT_DESCRIPTION]
        
        # support_server_invite_url
        self.support_server_invite_url = data.get(JSON_KEY_BOT_INFO_SUPPORT_SERVER_INVITE_URL, None)
        
        # tags
        self.tags = tuple(sorted(data[JSON_KEY_BOT_INFO_TAG_ARRAY]))
        
        # upvotes
        self.upvotes = data[JSON_KEY_BOT_INFO_UPVOTES]
        
        # upvotes_monthly
        self.upvotes_monthly = data[JSON_KEY_BOT_INFO_UPVOTES_MONTHLY]
        
        # vanity_url
        self.vanity_url = data.get(JSON_KEY_BOT_INFO_VANITY_URL, None)
        
        # website_url
        self.website_url = data.get(JSON_KEY_BOT_INFO_WEBSITE_URL, None)
        
        return self
    
    
    def __repr__(self):
        """Returns the bot info's representation."""
        return f'<{self.__class__.__name__} id={self.id} name={self.name}>'


class BotStats:
    """
    Contains a listed bot's stats.
    
    Attributes
    ----------
    guild_count : `int`
        The amount of guilds the bot is in. Defaults to `-1`.
    guild_count_per_shard : `tuple` of `int`
        The amount of guilds per shards. Can be empty.
    shard_count : `int`
        The amount of shards the bot has. Defaults to `-1`.
    shard_id : `int`
        The shard ID to post as (?). Defaults to `-1`.
    """
    __slots__ = ('guild_count', 'guild_count_per_shard', 'shard_count', 'shard_id')
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new bot stats instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Deserialized bot stats data.
        
        Returns
        -------
        self : ``BotStats``
        """
        self = object.__new__(cls)
        
        # guild_count
        self.guild_count = data.get(JSON_KEY_BOT_STATS_GUILD_COUNT, -1)
        
        # guild_count_per_shard
        try:
            guild_count_per_shard = data[JSON_KEY_BOT_STATS_GUILD_COUNT_PER_SHARD_ARRAY]
        except KeyError:
            guild_count_per_shard = ()
        self.guild_count_per_shard = guild_count_per_shard
        
        # shard_count
        self.shard_count = data.get(JSON_KEY_BOT_STATS_SHARD_COUNT, -1)
        
        # shard_id
        self.shard_id = data.get(JSON_KEY_BOT_STATS_SHARD_ID, -1)
        
        return self
    
    
    def __repr__(self):
        """Returns the bot stats' representation."""
        return f'<{self.__class__.__name__}>'


class UserInfo(metaclass=Slotted):
    """
    Represents a user.
    
    Attributes
    ----------
    avatar_hash : `int`
        The bot's avatar hash.
    avatar_type : ``IconType``
        The bot's avatar's type.
    banner_url : `None` or `str`
        Url for the user's banner image.
    bio : `None` or `str`
        The user's bio.
    color : ``Color``
        The custom color of the user.
    connections : ``UserConnections``
        Connections of the users to social networks.
    discriminator : `int`
        The user's discriminator.
    id : `int`
        The user's identifier.
    is_admin : `bool`
        Whether the user is an admin.
    is_certified_developer : `bool`
        Whether the user is a certified developer.
    is_moderator : `bool`
        Whether the user is moderator.
    is_supporter : `bool`
        Whether the user is a supporter.
    is_website_moderator : `bool`
        Whether the user is a website moderator.
    name : `str`
        The user's name.
    """
    __slots__ = ('banner_url', 'bio', 'color', 'connections', 'discriminator', 'id', 'is_admin',
        'is_certified_developer', 'is_moderator', 'is_supporter', 'is_website_moderator', 'name')

    avatar = IconSlot(
        'avatar',
        JSON_KEY_USER_INFO_AVATAR_BASE64,
        module_urls.user_avatar_url,
        module_urls.user_avatar_url_as,
    )
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new user info instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Deserialized user info data.
        
        Returns
        -------
        self : ``UserInfo``
        """
        self = object.__new__(cls)
        
        # avatar_hash & avatar_type
        self._set_avatar(data)
        
        # banner_url
        self.banner_url = data.get(JSON_KEY_USER_INFO_BANNER_URL, None)
        
        # bio
        self.bio = data.get(JSON_KEY_USER_INFO_BIO, None)
        
        # color
        try:
            color = Color(data[JSON_KEY_USER_INFO_COLOR], base=16)
        except ValueError:
            color = Color()
        self.color = color
        
        # connections
        self.connections = UserConnections.from_data(data[JSON_KEY_USER_INFO_CONNECTIONS])
        
        # discriminator
        self.discriminator = int(data[JSON_KEY_USER_INFO_DISCRIMINATOR_STRING])
        
        # id
        self.id = int(data[JSON_KEY_USER_INFO_ID])
        
        # is_admin
        self.is_admin = data[JSON_KEY_USER_INFO_IS_ADMIN]
        
        # is_certified_developer
        self.is_certified_developer = data[JSON_KEY_USER_INFO_IS_CERTIFIED_DEVELOPER]
        
        # is_moderator
        self.is_moderator = data[JSON_KEY_USER_INFO_IS_MODERATOR]
        
        # is_supporter
        self.is_supporter = data[JSON_KEY_USER_INFO_IS_SUPPORTER]
        
        # is_website_moderator
        self.is_website_moderator = data[JSON_KEY_USER_INFO_IS_WEBSITE_MODERATOR]
        
        # name
        self.name = data[JSON_KEY_USER_INFO_NAME]
        
        return self
    
    
    def __repr__(self):
        """Returns the bot info's representation."""
        return f'<{self.__class__.__name__} id={self.id} name={self.name}>'


class UserConnections:
    """
    Represents an user's connections.
    
    Attributes
    ----------
    github : `None` or `str`
        The github user name of the user.
    instagram : `None` or `str`
        The instagram user name of the user.
    reddit : `None` or `str`
        The reddit user name of the user.
    twitter : `None` or `str`
        The twitter user name of the user.
    youtube : `None` or `str`
        The youtube user name of the user.
    """
    __slots__ = ('github', 'instagram', 'reddit', 'twitter', 'youtube')
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new user info instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Deserialized user info data.
        
        Returns
        -------
        self : ``UserInfo``
        """
        self = object.__new__(cls)
        
        # github
        self.github = data.get(JSON_KEY_USER_INFO_CONNECTION_GITHUB, None)
        
        # instagram
        self.instagram = data.get(JSON_KEY_USER_INFO_CONNECTION_INSTAGRAM, None)
        
        # reddit
        self.reddit = data.get(JSON_KEY_USER_INFO_CONNECTION_REDDIT, None)
        
        # twitter
        self.twitter = data.get(JSON_KEY_USER_INFO_CONNECTION_TWITTER, None)
        
        # youtube
        self.youtube = data.get(JSON_KEY_USER_INFO_CONNECTION_YOUTUBE, None)
        
        return self
    
    
    def __repr__(self):
        """Returns the bot info's representation."""
        return f'<{self.__class__.__name__}>'


class BriefUserInfo(metaclass=Slotted):
    """
    Represents a user, containing only brief fields.
    
    Attributes
    ----------
    avatar_hash : `int`
        The bot's avatar hash.
    avatar_type : ``IconType``
        The bot's avatar's type.
    discriminator : `int`
        The user's discriminator.
    id : `int`
        The user's identifier.
    name : `str`
        The user's name.
    """
    __slots__ = ('discriminator', 'id', 'name')

    avatar = IconSlot(
        'avatar',
        JSON_KEY_USER_INFO_AVATAR_BASE64,
        module_urls.user_avatar_url,
        module_urls.user_avatar_url_as,
    )
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new user info instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Deserialized user info data.
        
        Returns
        -------
        self : ``UserInfo``
        """
        self = object.__new__(cls)
        
        # avatar_hash & avatar_type
        self._set_avatar(data)
        
        # discriminator
        self.discriminator = int(data[JSON_KEY_USER_INFO_DISCRIMINATOR_STRING])
        
        # id
        self.id = int(data[JSON_KEY_USER_INFO_ID])
        
        # name
        self.name = data[JSON_KEY_USER_INFO_NAME]
        
        return self
    
    
    def __repr__(self):
        """Returns the bot info's representation."""
        return f'<{self.__class__.__name__} id={self.id} name={self.name}>'


class BotsQueryResult:
    """
    Represents a get bots query's result.
    
    Attributes
    ----------
    count : `int`
        The number of received bots.
    limit : `int`
        The limit used.
    offset : `int`
        The off set used.
    results : `None` or `tuple` of ``BotInfo``
        The matched bots.
    total : `int`
        The total number of bots matching the query.
    """
    __slots__ = ('count', 'limit', 'offset', 'results', 'total')
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new bots query result instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Deserialized bots query result data.
        
        Returns
        -------
        self : ``BotsQueryResult``
        """
        self = object.__new__(cls)
        
        # count
        self.count = data[JSON_KEY_BOTS_QUERY_RESULT_COUNT]
        
        # limit
        self.limit = data[JSON_KEY_BOTS_QUERY_RESULT_LIMIT]
        
        # offset
        self.offset = data[JSON_KEY_BOTS_QUERY_RESULT_OFFSET]
        
        # results
        bot_info_datas = data[JSON_KEY_BOTS_QUERY_RESULT_RESULTS]
        if bot_info_datas:
            results = tuple(BotInfo.from_data(bot_info_data) for bot_info_data in bot_info_datas)
        else:
            results = None
        self.results = results
        
        # total
        self.total = data[JSON_KEY_BOTS_QUERY_RESULT_TOTAL]
        
        return self
    
    def __repr__(self):
        """Returns the bot info's representation."""
        return f'<{self.__class__.__name__} count={self.count} total={self.total}>'
    
    def __len__(self):
        """Returns the length of the query result"""
        return self.count
    
    def __iter__(self):
        """Iterates over the query result."""
        results = self.results
        if (results is not None):
            yield from results
    
    def __reversed__(self):
        """Reversed iterates over the query result."""
        results = self.results
        if (results is not None):
            yield from reversed(results)
    
    def __getitem__(self, index):
        """Gets the elements of the query result."""
        results = self.results
        if results is None:
            if isinstance(index, slice):
                return ()
            
            raise IndexError(index)
        
        return results[index]


class VoteBase:
    """
    Base class for vote instances received with webhook.
    
    Attributes
    ----------
    query : `str`
        Query used to redirect to vote.
    type : `str`
        The vote's type. Can be either `'upvote'` or `'test'`.
    user_id : `int`
        The user's identifier, who voted.
    """
    __slots__ = ('query', 'type', 'user_id')
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new vote instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Deserialized vote webhook data.
        
        Returns
        -------
        self : ``VoteBase``
        """
        self = object.__new__(cls)
        
        # query
        self.query = data[JSON_KEY_VOTE_BASE_QUERY]
        
        # type
        self.type = data[JSON_KEY_VOTE_BASE_TYPE]
        
        # user_id
        self.user_id = int(data[JSON_KEY_VOTE_BASE_USER_ID])
        
        return self
    
    def __repr__(self):
        """Returns the vote's representation."""
        return f'<{self.__class__.__name__} user_id={self.user_id!r}>'


class BotVote(VoteBase):
    """
    Bot vote received with webhook.
    
    Attributes
    ----------
    query : `str`
        Query used to redirect to vote.
    type : `str`
        The vote's type. Can be either `'upvote'` or `'test'`.
    user_id : `int`
        The user's identifier, who voted.
    bot_id : `int`
        The bot's identifier.
    is_weekend : `bool`
        Whether the vote was done on a weekend.
    """
    __slots__ = ('bot_id', 'is_weekend')

    @classmethod
    @copy_docs(VoteBase.from_data)
    def from_data(cls, data):
        self = super(BotVote, cls).from_data(data)
        
        self.bot_id = int(data[JSON_KEY_BOT_VOTE_BOT_ID])
        
        self.is_weekend = data.get(JSON_KEY_BOT_VOTE_IS_WEEKEND, False)
        
        return self
    
    @copy_docs(VoteBase.__repr__)
    def __repr__(self):
        return f'<{self.__class__.__name__} user_id={self.user_id!r} bot_id={self.bot_id!r}>'


class GuildVote(VoteBase):
    """
    Guild vote received with webhook.
    
    Attributes
    ----------
    query : `str`
        Query used to redirect to vote.
    type : `str`
        The vote's type. Can be either `'upvote'` or `'test'`.
    user_id : `int`
        The user's identifier, who voted.
    guild_id : `int`
        The guild's identifier.
    """
    __slots__ = ('guild_id',)

    @classmethod
    @copy_docs(VoteBase.from_data)
    def from_data(cls, data):
        self = super(GuildVote, cls).from_data(data)
        
        self.guild_id = int(data[JSON_KEY_GUILD_VOTE_GUILD_ID])
        
        return self
    
    @copy_docs(VoteBase.__repr__)
    def __repr__(self):
        return f'<{self.__class__.__name__} user_id={self.user_id!r} guild_id={self.guild_id!r}>'



__all__ = ('bind',)

from scarletio import WeakKeyDictionary, copy_docs

def bind(bind_to, bind_with, name):
    """
    A cool hata features, that lets you bind object to existing one.
    
    Parameters
    ----------
    bind_to : `type`
        The type to bind to.
    bind_with : `type`
        The type to bind with.
    name : `str`
        The name of the binding.
    
    Raises
    ------
    TypeError
        - If `bind_to` is not a type.
        - If `bind_to` instances do not support weakreferencing.
        - If `bind_to` has already an attribute named `name`.
    
    Examples
    --------
    Generic binder:
    ```py
    from hata import ClientUserBase, bind
    
    class Inventory(object):
        def __init__(self, parent_self):
            self.user = parent_self
            self.inv = []
        
        def add_item(self, name):
            self.inv.append(name)
        
        def give_user(self, item, recipient):
            del self.inv[self.inv.index(item)]
            recipient.inventory.add_item(item)
    
    bind(ClientUserBase, Inventory, 'inventory')
    
    # Usage:
    user.inventory.add_item('cake')
    ```
    
    Descriptor binder:
    ```py
    from hata import ClientUserBase, bind
    
    class InventorySize(object):
        def __init__(self, parent_self):
            self.value = 10
        
        def __get__(self, instance, type_):
            return self.value
        
        def __set__(self, instance, value):
            self.value = value
    
    bind(ClientUserBase, InventorySize, 'inventory_size')
    
    # Usage:
    user.inventory_size += 10
    ```
    Descriptor binders are familiar ot normal descriptors, except, they are created per object and not per class.
    
    """
    if not isinstance(bind_to, type):
        raise TypeError(f'`bind_to` to must be a type, got {bind_to.__class__.__name__}; {bind_to!r}.')
    
    name_space = bind_to.__dict__
    if ('__weakref__' in name_space) or ('__slots__' not in name_space):
        raise TypeError(f'`bind_to` instances must support weakreferencing, got {bind_to!r}.')
    
    if hasattr(bind_to, name):
        raise TypeError(f'`bind_to` already has an attribute named, as bind_to=`{bind_to!r}`, name={name!r}.')
    
    if (getattr(bind_with, '__get__', None) is not None):
        binder = DescriptorObjectBinder(name, bind_with)
    else:
        binder = GenericObjectBinder(name, bind_with)
    
    setattr(bind_to, name, binder)


class ObjectBinderBase:
    """
    Object binder.
    
    Attributes
    ----------
    cache : ``WeakKeyDictionary``
        The binding relation mapping.
    name : `str`
        The filed's name.
    type : `type`
        The binded class.
    """
    __slots__ = ('cache', 'name', 'type',)
    
    def __new__(cls, name, type_):
        """
        Creates a new object binder instance.
        
        Parameters
        ----------
        name : `str`
            The filed's name.
        type_ : `type`
            The binded class.
        """
        self = object.__new__(cls)
        self.cache = WeakKeyDictionary()
        self.name = name
        self.type = type_
        return self
    
    
    def __get__(self, instance, type_):
        """
        Gets the field's value.
        
        Parameters
        ----------
        instance : `Any`
            The instance itself.
        type_ : `type`
            The instance's type.
        
        Returns
        -------
        value : `self` / `Any`
            Returns itself if called from class.
        """
        if (instance is None):
            return self
    
    
    def __set__(self, instance, value):
        """
        Sets the value to the instance.
        
        Parameters
        ----------
        instance : `Any`
            The parent instance to set the value to.
        value : `Any`
            The value to set.
        
        Raises
        ------
        AttributeError
            Read only attribute.
        """
        raise AttributeError(f'Read only attribute: {self.name!r}.')
    
    
    def __delete__(self, instance):
        """
        Deletes the value of the instance.
        
        Parameters
        ----------
        instance : `Any`
            The parent instance to delete the attribute of.
        
        Raises
        ------
        AttributeError
            Cannot delete attribute.
        """
        raise AttributeError(f'Cannot delete attribute: {self.name!r}.')
    

class GenericObjectBinder(ObjectBinderBase):
    __slots__ = ()
    
    @copy_docs(ObjectBinderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        try:
            bind_object = self.cache[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.cache[instance] = bind_object
        
        return bind_object


class DescriptorObjectBinder(ObjectBinderBase):
    """
    Object binder for descriptors-likes.
    
    Attributes
    ----------
    cache : ``WeakKeyDictionary``
        The binding relation mapping.
    name : `str`
        The filed's name.
    type : `type`
        The binded class.
    fdel : `None` or `Any`
        Deleter function if applicable.
    fget : `None` or `Any`
        Getter function if applicable.
    fset : `None` or `Any`
        Setter function if applicable.
    """
    __slots__ = ('fdel', 'fget', 'fset')
    
    @copy_docs(ObjectBinderBase.__new__)
    def __new__(cls, name, type_):
        self = ObjectBinderBase.__new__(cls, name, type_)
        self.fdel = getattr(type_, '__delete__', None)
        self.fget = getattr(type_, '__get__')
        self.fset = getattr(type_, '__set__', None)
        return self
    
    
    @copy_docs(ObjectBinderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        try:
            bind_object = self.cache[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.cache[instance] = bind_object
        
        return self.fget(bind_object, instance, type_)
    
    
    @copy_docs(ObjectBinderBase.__get__)
    def __set__(self, instance, value):
        fset = self.fset
        if fset is None:
            raise AttributeError(f'Read only attribute: {self.name!r}.')
        
        try:
            bind_object = self.cache[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.cache[instance] = bind_object
        
        return fset(bind_object, instance, value)
    
    
    @copy_docs(ObjectBinderBase.__delete__)
    def __delete__(self, instance):
        fdel = self.fdel
        if fdel is None:
            raise AttributeError(f'Cannot delete attribute: {self.name!r}.')
        
        try:
            bind_object = self.cache[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.cache[instance] = bind_object
        
        return fdel(bind_object, instance)
